require("stuff/preamble")

local PrePro = {}

local Buffer = lib.stuff.Class("PrePro.Buffer")

function Buffer:__call(...)
	for _, i in ipairs{...} do
		table.insert(self, tostring(i))
	end
end

function Buffer:__tostring()
	return table.concat(self)
end

PrePro.Buffer = Buffer

function PrePro.transform(code)	
	local res = Buffer()
	local buf = Buffer()
	
	local function wrapUp()
		local s = tostring(buf)
		buf = Buffer()
		return s
	end
	
	local escapes = {
		["\""] = "\\\"",
		["\r"] = "\\r",
		["\n"] = "\\n",
		["\\"] = "\\\\",
	}
	
	local function escapeChar(c)
		return escapes[c] or c
	end
	
	--[[
	States:
		"code", "metacode", "metaexp", "newline"
	]]
	
	local state = "newline"
	
	local i = 1
	while i <= #code do
		local pair = code:sub(i, i+1)
		local first = code:sub(i, i)
		local second = code:sub(i+1, i+1)
		
		if state == "newline" then
			if first == "$" then
				state = "metacode"
				i = i + 1
			elseif first == "\n" then
				res("emit(\"\\n\")\n")
				i = i + 1
			else
				state = "code"
			end
		elseif state == "code" then
			if first == "\n" then
				res("emit(\"", wrapUp(), "\\n\")\n")
				state = "newline"
				i = i + 1
			elseif pair == "$(" then
				if #buf > 0 then
					res("emit(\"", wrapUp(), "\") ")
				end
				state = "metaexp"
				i = i + 2
			else
				buf(escapeChar(first))
				i = i + 1
			end
		elseif state == "metacode" then
			if first == "\n" then
				res(wrapUp(), "\n")
				state = "newline"
				i = i + 1
			else
				buf(first)
				i = i + 1
			end
		elseif state == "metaexp" then
			if pair == ")$" then
				res("emit(", wrapUp(), ") ")
				state = "code"
				i = i + 2
			else
				buf(first)
				i = i + 1
			end
		else
			print("error")
			i = i + 1
		end
	end
	
	return tostring(res)
end

function PrePro.loadString(s)
	local t = PrePro.transform(s)
	--~ PrePro.writeFile("test_PrePro_debug.lua", t)
	return assert(loadstring(t))
end

function PrePro.loadFile(path)
	local f = io.open(path, "rb")
	local res = PrePro.loadString(f:read("*all"))
	f:close()
	return res
end

function PrePro.writeFile(path, s)
	local f = io.open(path, "wb")
	f:write(s)
	f:close()
end

function PrePro.makeEnv()
	local emit = Buffer()
	local isRequired = {}
	local required = {}
	local env = {emit = emit}
		
	function env.preRequire(path)
		local f = PrePro.loadFile(path)
		setfenv(f, env)
		required[path] = f()
		isRequired[path] = true
		return required[path]
	end
	
	function env.compile()
		--~ PrePro.writeFile("test_PrePro_debug.lua", tostring(emit))
		--~ return require("test_PrePro_debug")
		return assert(loadstring(tostring(emit)))()
	end
	
	setmetatable(env, {__index = _G})
	
	return env
end

return PrePro
